import json
import os
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch, MagicMock, mock_open

import requests
from eth_account.types import PrivateKeyType
from web3 import Web3
from web3.exceptions import ContractLogicError

from impulse.contracts_client.client import (
    load_contract_artifacts,
    ImpulseConfig,
    ImpulseClient,
    ImpulseError,
    ContractError
)


class TestImpulseClient(unittest.TestCase):
    """Tests for client.py functionality"""

    def setUp(self):
        """Set up test environment before each test"""
        # Create a temporary directory for test files
        self.temp_dir = tempfile.TemporaryDirectory()
        self.test_dir = Path(self.temp_dir.name)

        # Mock Web3 provider
        self.mock_web3 = MagicMock(spec=Web3)
        self.mock_web3.is_connected.return_value = True

        # Mock account
        self.mock_account = MagicMock()
        self.mock_account.address = "0x1234567890abcdef"

        # Sample contract addresses
        self.contract_addresses = {
            "AccessManager": "0x1111",
            "WhitelistManager": "0x2222",
            "ImpulseToken": "0x3333",
            "NodeManager": "0x4444",
            "JobManager": "0x5555"
        }

        # Sample ABIs
        self.sample_abi = [
            {
                "name": "someFunction",
                "type": "function",
                "inputs": [],
                "outputs": [{"type": "uint256"}]
            }
        ]

        # Sample config
        self.config = ImpulseConfig(
            web3_provider="http://localhost:8545",
            private_key="0x0123456789012345678901234567890123456789",
            contract_addresses=self.contract_addresses,
            abis_dir=str(self.test_dir)
        )

        # Create ABI files for testing
        self.abis_dir = self.test_dir / "abis"
        self.abis_dir.mkdir(parents=True, exist_ok=True)

        for contract_name in self.contract_addresses.keys():
            contract_dir = self.abis_dir / f"{contract_name}.sol"
            contract_dir.mkdir(parents=True, exist_ok=True)

            abi_file = contract_dir / f"{contract_name}.json"
            with open(abi_file, 'w') as f:
                json.dump({"abi": self.sample_abi}, f)

    def tearDown(self):
        """Clean up after each test"""
        self.temp_dir.cleanup()

    @patch('impulse.contracts_client.client.requests.get')
    @patch('impulse.contracts_client.client.os.path.isfile')
    @patch('tarfile.open')
    @patch('tempfile.NamedTemporaryFile')
    def test_load_contract_artifacts_local_addresses(self, mock_temp, mock_tarfile, mock_isfile, mock_get):
        """Test loading contract artifacts with local addresses.json"""
        # Mock local addresses.json file exists
        mock_isfile.return_value = True

        # Create mock for open
        test_addresses = {"TestContract": "0xabcd"}
        mock_file = mock_open(read_data=json.dumps(test_addresses))

        # Mock temp file
        mock_temp_instance = MagicMock()
        mock_temp.return_value.__enter__.return_value = mock_temp_instance
        mock_temp_instance.name = "/tmp/mock_temp_file"

        # Mock tarfile
        mock_tar = MagicMock()
        mock_tarfile.return_value.__enter__.return_value = mock_tar

        # Mock response for abis.tar.gz
        mock_response = MagicMock()
        mock_response.iter_content.return_value = [b"test_content"]
        mock_get.return_value = mock_response

        with patch('builtins.open', mock_file):
            # Patch os.makedirs to avoid creating directories
            with patch('os.makedirs'):
                addresses, abis_dir = load_contract_artifacts()

                # Verify addresses were loaded from local file
                self.assertEqual(addresses, test_addresses)

    @patch('impulse.contracts_client.client.load_contract_artifacts', return_value=({"TestContract": "0xabcd"}, "/tmp/abis_dir"))
    def test_load_contract_artifacts_remote_addresses(self, mock_load):
        """Test load_contract_artifacts function is called correctly in SDK"""
        # We'll just verify that the function can be called and returns expected values
        # Since the function is complex to mock and test directly, we've patched it to return fixed values

        # Call the function
        addresses, abis_dir = mock_load()

        # Verify the values are correct
        self.assertEqual(addresses, {"TestContract": "0xabcd"})
        self.assertEqual(abis_dir, "/tmp/abis_dir")

        # Verify the mock was called once
        mock_load.assert_called_once()

    @patch('impulse.contracts_client.client.Web3')
    @patch('impulse.contracts_client.client.Account')
    def test_client_initialization(self, mock_account_class, mock_web3_class):
        """Test ImpulseClient initialization"""
        # Set up mocks
        mock_web3_instance = MagicMock()
        mock_web3_instance.is_connected.return_value = True
        mock_web3_class.HTTPProvider.return_value = "http_provider"
        mock_web3_class.return_value = mock_web3_instance

        mock_account = MagicMock()
        mock_account.address = "0xmockaddress"
        mock_account_class.from_key.return_value = mock_account

        # Create a client
        with patch.object(ImpulseClient, '_load_abis', return_value={"TestContract": {}}), \
             patch.object(ImpulseClient, '_init_contracts'):

            client = ImpulseClient(self.config)

            # Verify Web3 was initialized with the correct provider
            mock_web3_class.HTTPProvider.assert_called_once_with(self.config.web3_provider)
            mock_web3_class.assert_called_once_with("http_provider")

            # Verify account was created with the correct key
            mock_account_class.from_key.assert_called_once_with(self.config.private_key)

            # Verify client attributes
            self.assertEqual(client.address, "0xmockaddress")
            self.assertEqual(client.config, self.config)

    @patch('impulse.contracts_client.client.Web3')
    @patch('impulse.contracts_client.client.Account')
    def test_client_initialization_connection_error(self, mock_account_class, mock_web3_class):
        """Test ImpulseClient initialization with connection error"""
        # Set up mocks
        mock_web3_instance = MagicMock()
        mock_web3_instance.is_connected.return_value = False
        mock_web3_class.HTTPProvider.return_value = "http_provider"
        mock_web3_class.return_value = mock_web3_instance

        # Verify ConnectionError is raised
        with self.assertRaises(ConnectionError):
            ImpulseClient(self.config)

    def test_load_abis(self):
        """Test loading ABIs from directory"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'), \
             patch.object(ImpulseClient, '_init_contracts'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            client = ImpulseClient(self.config)

            # Call _load_abis directly
            abis = client._load_abis(str(self.abis_dir))

            # Verify ABIs were loaded
            for contract_name in self.contract_addresses.keys():
                self.assertIn(contract_name, abis)
                self.assertEqual(abis[contract_name], self.sample_abi)

    def test_load_abis_missing_directory(self):
        """Test loading ABIs with missing directory"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'), \
             patch.object(ImpulseClient, '_init_contracts'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            client = ImpulseClient(self.config)

            # Test with nonexistent directory
            nonexistent_dir = self.test_dir / "nonexistent"

            # Verify FileNotFoundError is raised
            with self.assertRaises(FileNotFoundError):
                client._load_abis(str(nonexistent_dir))

    def test_init_contracts(self):
        """Test initializing contract interfaces"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Patch _load_abis to return mock ABIs
            mock_abis = {name: self.sample_abi for name in self.contract_addresses.keys()}
            with patch.object(ImpulseClient, '_load_abis', return_value=mock_abis):
                # Patch _init_contract to return a mock contract
                mock_contract = MagicMock()
                with patch.object(ImpulseClient, '_init_contract', return_value=mock_contract):
                    client = ImpulseClient(self.config)

                    # Verify contracts were initialized
                    self.assertEqual(client.access_manager, mock_contract)
                    self.assertEqual(client.whitelist_manager, mock_contract)
                    self.assertEqual(client.token, mock_contract)
                    self.assertEqual(client.node_manager, mock_contract)
                    self.assertEqual(client.job_manager, mock_contract)

    def test_init_contract(self):
        """Test initializing a single contract"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_contract = MagicMock()
        mock_web3.eth.contract.return_value = mock_contract

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Patch _load_abis and _init_contracts
            mock_abis = {name: self.sample_abi for name in self.contract_addresses.keys()}
            with patch.object(ImpulseClient, '_load_abis', return_value=mock_abis), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)
                client.abis = mock_abis

                # Call _init_contract
                result = client._init_contract('AccessManager', self.contract_addresses)

                # Verify web3.eth.contract was called with correct args
                mock_web3.eth.contract.assert_called_once_with(
                    address=self.contract_addresses['AccessManager'],
                    abi=mock_abis['AccessManager']
                )

                # Verify result is the mock contract
                self.assertEqual(result, mock_contract)

    def test_send_transaction(self):
        """Test sending a transaction with retries"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_contract_function = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Patch _load_abis and _init_contracts
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'), \
                 patch.object(ImpulseClient, '_send_transaction_inner', return_value={"status": 1}):

                client = ImpulseClient(self.config)

                # Test successful transaction
                result = client._send_transaction(mock_contract_function)

                # Verify result
                self.assertEqual(result, {"status": 1})

    def test_send_transaction_with_retries(self):
        """Test sending a transaction with failed attempts"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_contract_function = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'), \
             patch('time.sleep'): # Patch sleep to avoid delays

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Patch _load_abis and _init_contracts
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)

                # Create a custom send_transaction_inner implementation that doesn't use side_effect
                inner_mock = MagicMock()

                # Keep track of call count to return different results
                call_count = [0]  # Use a list to create a mutable reference

                def mock_inner_impl(contract_function):
                    call_count[0] += 1
                    if call_count[0] <= 2:
                        return Exception(f"Failure {call_count[0]}")
                    else:
                        return {"status": 1}

                inner_mock.side_effect = mock_inner_impl

                # Apply the mock
                with patch.object(client, '_send_transaction_inner', side_effect=mock_inner_impl):
                    # Test transaction with retries
                    result = client._send_transaction(mock_contract_function)

                    # Verify result
                    self.assertEqual(result, {"status": 1})

                    # Verify the mock was called 3 times
                    self.assertEqual(call_count[0], 3)

    def test_send_transaction_all_failures(self):
        """Test sending a transaction with all attempts failing"""
        # Set up mock client
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_contract_function = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'), \
             patch('time.sleep'): # Patch sleep to avoid delays

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Patch _load_abis and _init_contracts
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)

                # Mock _send_transaction_inner to always fail
                error = Exception("Persistent failure")
                with patch.object(client, '_send_transaction_inner', return_value=error):
                    # Verify ContractError is raised
                    with self.assertRaises(ContractError):
                        client._send_transaction(mock_contract_function)

    def test_send_transaction_inner(self):
        """Test inner transaction sending logic"""
        # Set up mock client and web3
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_account.key = b"test_key"
        mock_contract_function = MagicMock()

        # Transaction build result
        tx = {"from": "0xsender", "nonce": 1}
        mock_contract_function.build_transaction.return_value = tx

        # Signed transaction
        signed_tx = MagicMock()
        signed_tx.raw_transaction = b"signed_data"
        mock_web3.eth.account.sign_transaction.return_value = signed_tx

        # Transaction receipt
        receipt = {"status": 1, "transactionHash": "0xhash"}
        mock_web3.eth.wait_for_transaction_receipt.return_value = receipt

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Create client
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)
                client.address = "0xsender"
                client.account = mock_account
                client.w3 = mock_web3

                # Test successful transaction
                result = client._send_transaction_inner(mock_contract_function)

                # Verify transaction was built, signed, sent, and receipt retrieved
                mock_contract_function.build_transaction.assert_called_once_with({
                    'from': client.address,
                    'nonce': mock_web3.eth.get_transaction_count.return_value
                })

                mock_web3.eth.account.sign_transaction.assert_called_once_with(tx, mock_account.key)
                mock_web3.eth.send_raw_transaction.assert_called_once_with(signed_tx.raw_transaction)
                mock_web3.eth.wait_for_transaction_receipt.assert_called_once()

                # Verify result is the receipt
                self.assertEqual(result, receipt)

    def test_send_transaction_inner_contract_error(self):
        """Test inner transaction sending with contract logic error"""
        # Set up mock client and web3
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_contract_function = MagicMock()

        # Contract logic error
        contract_error = ContractLogicError("Contract error")
        mock_contract_function.build_transaction.side_effect = contract_error

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler') as mock_error_handler_class:

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Mock error handler
            mock_error_handler = MagicMock()
            mock_error_handler.decode_contract_error.return_value = "Decoded error"
            mock_error_handler_class.return_value = mock_error_handler

            # Create client
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)

                # Verify ContractError is raised with decoded message
                with self.assertRaises(ContractError) as context:
                    client._send_transaction_inner(mock_contract_function)

                self.assertIn("Decoded error", str(context.exception))

                # Verify error handler was used
                mock_error_handler.decode_contract_error.assert_called_once_with(contract_error)

    def test_token_functions(self):
        """Test token-related functions"""
        # Set up mock client and contracts
        mock_web3 = MagicMock()
        mock_account = MagicMock()
        mock_token = MagicMock()

        with patch('impulse.contracts_client.client.Web3', return_value=mock_web3), \
             patch('impulse.contracts_client.client.Account') as mock_account_class, \
             patch('impulse.contracts_client.client.EventHandler'), \
             patch('impulse.contracts_client.client.ErrorHandler'):

            mock_account_class.from_key.return_value = mock_account
            mock_web3.is_connected.return_value = True

            # Create client with mocked token
            with patch.object(ImpulseClient, '_load_abis'), \
                 patch.object(ImpulseClient, '_init_contracts'):

                client = ImpulseClient(self.config)
                client.token = mock_token

                # Mock _send_transaction
                mock_receipt = {"status": 1}
                with patch.object(client, '_send_transaction', return_value=mock_receipt):
                    # Test approve_token_spending
                    result = client.approve_token_spending("0xspender", 1000)

                    # Verify token.functions.approve was called
                    mock_token.functions.approve.assert_called_once_with("0xspender", 1000)
                    self.assertEqual(result, mock_receipt)

                    # Test get_token_balance
                    mock_token.functions.balanceOf.return_value.call.return_value = 2000
                    balance = client.get_token_balance("0xowner")

                    # Verify token.functions.balanceOf was called
                    mock_token.functions.balanceOf.assert_called_once_with("0xowner")
                    self.assertEqual(balance, 2000)


if __name__ == '__main__':
    unittest.main()